@@ -8,8 +8,9 @@ from libnacl.version import __version__
 # Import python libs
 import ctypes
 import sys
+import os
 
-__SONAMES = (17, 13, 10, 5, 4)
+__SONAMES = (18, 17, 13, 10, 5, 4)
 
 
 def _get_nacl():
@@ -68,6 +69,12 @@ def _get_nacl():
                 )
             except OSError:
                 pass
+        try:
+            # fall back to shipped libsodium, trust os version first
+            libpath = os.path.join(os.path.dirname(__file__), 'libsodium.so')
+            return ctypes.cdll.LoadLibrary(libpath)
+        except OSError:
+            pass
         msg = 'Could not locate nacl lib, searched for libsodium.so, '
         for soname_ver in __SONAMES:
             msg += 'libsodium.so.{0}, '.format(soname_ver)
@@ -89,11 +96,13 @@ if sodium_init() < 0:
 
 # Define constants
 crypto_box_SECRETKEYBYTES = nacl.crypto_box_secretkeybytes()
+crypto_box_SEEDBYTES = nacl.crypto_box_seedbytes()
 crypto_box_PUBLICKEYBYTES = nacl.crypto_box_publickeybytes()
 crypto_box_NONCEBYTES = nacl.crypto_box_noncebytes()
 crypto_box_ZEROBYTES = nacl.crypto_box_zerobytes()
 crypto_box_BOXZEROBYTES = nacl.crypto_box_boxzerobytes()
 crypto_box_BEFORENMBYTES = nacl.crypto_box_beforenmbytes()
+crypto_box_SEALBYTES = nacl.crypto_box_sealbytes()
 crypto_scalarmult_BYTES = nacl.crypto_scalarmult_bytes()
 crypto_scalarmult_SCALARBYTES = nacl.crypto_scalarmult_scalarbytes()
 crypto_sign_BYTES = nacl.crypto_sign_bytes()
@@ -215,7 +224,7 @@ def crypto_box_afternm(msg, nonce, k):
     ctxt = ctypes.create_string_buffer(len(pad))
     ret = nacl.crypto_box_afternm(ctxt, pad, ctypes.c_ulonglong(len(pad)), nonce, k)
     if ret:
-        raise ValueError('Unable to encrypt messsage')
+        raise CryptError('Unable to encrypt messsage')
     return ctxt.raw[crypto_box_BOXZEROBYTES:]
 
 
@@ -236,9 +245,43 @@ def crypto_box_open_afternm(ctxt, nonce, k):
             nonce,
             k)
     if ret:
-        raise ValueError('unable to decrypt message')
+        raise CryptError('unable to decrypt message')
     return msg.raw[crypto_box_ZEROBYTES:]
 
+def crypto_box_seal(msg, pk):
+    '''
+    Using a public key to encrypt the given message. The identity of the sender cannot be verified.
+
+    enc_msg = nacl.crypto_box_seal('secret message', <public key string>)
+    '''
+    if len(pk) != crypto_box_PUBLICKEYBYTES:
+        raise ValueError('Invalid public key')
+    if not isinstance(msg, bytes):
+        raise TypeError('Message must be bytes')
+
+    c = ctypes.create_string_buffer(len(msg) + crypto_box_SEALBYTES)
+    ret = nacl.crypto_box_seal(c, msg, ctypes.c_ulonglong(len(msg)), pk)
+    if ret:
+        raise CryptError('Unable to encrypt message')
+    return c.raw
+
+def crypto_box_seal_open(ctxt, pk, sk):
+    '''
+    Decrypts a message given the receiver's public and private key.
+    '''
+    if len(pk) != crypto_box_PUBLICKEYBYTES:
+        raise ValueError('Invalid public key')
+    if len(sk) != crypto_box_SECRETKEYBYTES:
+        raise ValueError('Invalid secret key')
+    if not isinstance(ctxt, bytes):
+        raise TypeError('Message must be bytes')
+    
+    c = ctypes.create_string_buffer(len(ctxt) - crypto_box_SEALBYTES)
+    ret = nacl.crypto_box_seal_open(c, ctxt, ctypes.c_ulonglong(len(ctxt)), pk, sk)
+    if ret:
+        raise CryptError('Unable to decrypt message')
+    return c.raw
+
 # Signing functions
 
 
@@ -273,7 +316,7 @@ def crypto_sign(msg, sk):
 
 def crypto_sign_seed_keypair(seed):
     '''
-    Computes and returns the secret adn verify keys from the given seed
+    Computes and returns the secret and verify keys from the given seed
     '''
     if len(seed) != crypto_sign_SEEDBYTES:
         raise ValueError('Invalid Seed')
@@ -593,6 +636,28 @@ def crypto_verify_64(string1, string2):
     '''
     return not nacl.crypto_verify_64(string1, string2)
 
+
+def bytes_eq(a, b):
+    '''
+    Compares two byte instances with one another. If `a` and `b` have
+    different lengths, return `False` immediately. Otherwise `a` and `b`
+    will be compared in constant time.
+
+    Return `True` in case `a` and `b` are equal. Otherwise `False`.
+
+    Raises :exc:`TypeError` in case `a` and `b` are not both of the type
+    :class:`bytes`.
+    '''
+    if not isinstance(a, bytes) or not isinstance(b, bytes):
+        raise TypeError('Both arguments must be bytes.')
+
+    len_a = len(a)
+    len_b = len(b)
+    if len_a != len_b:
+        return False
+
+    return nacl.sodium_memcmp(a, b, len_a) == 0
+
 # Random byte generation
 
 
@@ -670,3 +735,18 @@ def sodium_version_string():
     func = nacl.sodium_version_string
     func.restype = ctypes.c_char_p
     return func()
+
+
+def crypto_box_seed_keypair(seed):
+    '''
+    Computes and returns the public and secret keys from the given seed
+    '''
+    if len(seed) != crypto_box_SEEDBYTES:
+        raise ValueError('Invalid Seed')
+    pk = ctypes.create_string_buffer(crypto_box_PUBLICKEYBYTES)
+    sk = ctypes.create_string_buffer(crypto_box_SECRETKEYBYTES)
+
+    ret = nacl.crypto_box_seed_keypair(pk, sk, seed)
+    if ret:
+        raise CryptError('Failed to generate keypair from seed')
+    return (pk.raw, sk.raw)
